import os
from mako.template import Template
from pathlib import Path
from . import utils
import collections
from lupa import *

class KV:
    def __init__(self):
        self.k = None
        self.v = None


class AttrClass():
    def __init__(self):
        self.class_name = str.lower(self.__class__.__name__)
        self.on_init()

    def on_init(self):
        return


class PrimaryIndex(AttrClass):
    def on_init(self):
        return


class Index(AttrClass):
    def on_init(self):
        return


class MuIndex(AttrClass):
    def on_init(self):
        self.index_data = []
        self.funcName = ""
        return

    def addData(self, comp_name, comp_value):
        temp_t = KV()
        temp_t.k = comp_name
        temp_t.v = comp_value
        self.index_data.append(temp_t)
        return

    def setFuncName(self, name):
        self.funcName = name


class Event(AttrClass):
    def __init__(self):
        super().__init__()
        self.event = None
        self.target = None
        self.type = None
        self.priority = 0
        self.action = None

    def on_init(self):
        self.event = None
        return

    def init_event(self, event):
        self.event = event
        self.target = event.target
        self.type = event.type or 'ADDED'
        self.priority = event.priority or 0
        self.action = event.action

    def get_group_event(self):
        if self.type != "ALL":
            return "GroupEvent." + self.type
        else:
            return "GroupEvent.ADDED | GroupEvent.REMOVED"


class Component():
    def __init__(self, name):
        self.data = []
        self.single = False
        self.attr = []
        self.simple = True
        self.name = name[0].lower() + name[1:]
        self.tag = None
        self.has_attr = False
        self.Name = name[0].upper() + name[1:]
        self.event_target = None
        self.event_action = None
        self.event = []
        return

    def __cmp__(self, other):
        if self.name > other.name:
            return 1
        else:
            return -1

    def is_empty(self):
        return len(self.data) == 0

    def set_data(self, data):
        self.data = []
        self.simple = False
        for i in range(len(data)):
            p = data[i]
            sp_ret = p.split('@')
            self.parse_property_data(i, sp_ret[0])
            if len(sp_ret) > 1:
                sp_ret = sp_ret[1].split('|')
                for prop in sp_ret:
                    prop = prop.replace(' ', '')
                    self.parse_property_data_attr(prop, self.data[i][0])

    def parse_property_data(self, i, property):
        sp_ret = property.split(':')
        self.data.append([0, 0])
        self.data[i][0] = sp_ret[0].replace(' ', '')
        self.data[i][1] = sp_ret[1]
        return

    def parse_property_data_attr(self, attr, p_name):
        if attr == 'primaryIndex':
            temp = PrimaryIndex()
            temp.p_name = p_name
            self.attr.append(temp)
        elif attr == 'index':
            temp = Index()
            temp.p_name = p_name
            self.attr.append(temp)
        return

    def get_property(self, i, context):
        ret_str = self.data[i][1]
        if context.pre_tag:
            ret_str = ret_str.replace('${pre_tag}', context.pre_tag).replace(' ', '')
        return ret_str

    def set_single(self, single):
        self.single = single

    def set_event(self, events):
        for index in events:
            event = events[index]
            if not event.target:
                raise ("event must has target")
            e = Event()
            e.init_event(event)
            self.event.append(e)

    def add_attr(self, key, val):
        self.attr[key] = self['parser_attr' + key](val)

    def parser_attr_primaryIndex(self, attr):
        for at in attr:
            name = at.split(':')
            name.replace(' ', '')
            is_have = False
            for data in self.data:
                if data[0] == name:
                    is_have = True
                    break
            if not is_have:
                raise (' attr value not in comp')

    def parser_attr_Index(self, attr):
        for at in attr:
            name = at.split(':')
            name.replace(' ', '')
            is_have = False
            for data in self.data:
                if data[0] == name:
                    is_have = True
                    break
            if not is_have:
                raise (' attr value not in comp')

    def parser_attr_Event(self, attr):
        for at in attr:
            name = at.split(':')
            name.replace(' ', '')
            is_have = False
            for data in self.data:
                if data[0] == name:
                    is_have = True
                    break
            if not is_have:
                raise (' attr value not in comp')

        return attr

    def get_func_params(self, pre='', sep=', '):
        b = []

        for item in self.data:
            b.append(pre + item[0])
        if len(self.data) > 0:
            return ', ' + sep.join(b)
        return ''

    def set_tag(self, tags):
        self.tag = tags

    def add_event(self, event):
        if event.eventTarget is None:
            raise ("error")
        self.event.append(event)


class ContextData():
    def __init__(self, tag):
        self.components = []
        self.event_comps = []
        self.source = ""
        self.name = tag
        self.Name = tag[0].upper() + tag[1:]
        self.muIndex = []
        self.tag = tag
        self.pre_tag = ""
        self.simple_name = tag
        if '_' in self.tag:
            self.pre_tag = self.tag[0:2]
            self.simple_name = self.simple_name.replace(self.pre_tag, '')

    def setSource(self, source):
        self.source = source

    def setOutPut(self, output):
        self.output = output

    def setComponents(self, comps):
        self.components = comps

    def addComponent(self, comp):
        self.components.append(comp)

    def addEventComponent(self, comp):
        self.event_comps.append(comp)

    def check(self):
        for comp in self.components:
            if comp.tag is None:
                raise (comp.name + "tag is None")

    def addContextMuIndex(self, index):
        self.muIndex.append(index)
        return


class BaseParser:
    def __init__(self, config_file_path):
        self.lua = LuaRuntime()
        self.source = ""
        self.namespace = ""
        self.out_path = ""
        self.parser_tag = None
        self.tag = None
        self.component_path = ''
        self.contexts = collections.OrderedDict()
        self.service_path = Path("")
        self.mako_path = Path(utils.get_python_fiel_Path(__file__)) / '../mako'
        self.python_entitas_soure_path = Path(utils.get_python_fiel_Path(__file__)) / '../entitas'

        self.base_config_file = config_file_path
        self.base_config_path = Path(os.path.dirname(config_file_path))
        self.script_path = Path()

        self.on_init()
        self.load_base_config()
        self.load_context_config()

    def load_base_config(self):
        with open(self.base_config_file , 'r') as load_f:
            text = load_f.read()
            config = self.lua.compile(text)()
        self.source = config.source
        self.component_path = self.base_config_path / config.component_path

        self.out_path = self.base_config_path / config.output
        self.namespace = config.namespace
        self.tag = list(config.tag.values())
        self.service_path = self.base_config_path / config.service_path
        self.tag.sort()

        for tag in self.tag:
            context = ContextData(tag)
            context.setSource(self.source)
            context.setOutPut(self.out_path)
            self.contexts[tag] = context

    def load_context_config(self):
        return

    def on_init(self):
        return

    def get_files_by_path(self):
        return

    def render_mako(self, fime_name, mako_name, context):
        file_path = os.path.join(self.out_path, context.name + fime_name)
        file = utils.open_file(file_path + '.py', 'w')
        template = Template(filename=str(self.mako_path / mako_name),
                            module_directory=os.path.join(self.script_path, 'makoCache'))
        content = self.template_render(template, context)
        content = content.replace('\n', '')
        content = content.replace('\r\n', '')
        file.write(content)
        file.close()

    def template_render(self, template, context):
        return template.render(
            context_name=context.name,
            contexts=context,
            source_path=self.source)

    def generate_context(self):
        for key, context in self.contexts.items():
            self.render_mako("Context" , 'ecs_context.mako', context)
            file_name = os.path.join(self.out_path / "../Extension/Context", context.name + "Context.py" )
            if not os.path.exists(file_name):
                file = utils.open_file(file_name, 'w')
                file.write('''
from ...Generated.{0}Context import {0}Context as Context
class {0}Context(Context):
    def __init__(self):
        super().__init__()'''.format(context.name))
                file.close()

    def generate_entity(self):
        for key, context in self.contexts.items():
            self.render_mako("Entity" , 'ecs_entity.mako', context)
            file_name = os.path.join(self.out_path / "../Extension/Entity", context.name + "Entity.py" )
            if not os.path.exists(file_name):
                file = utils.open_file(file_name, 'w')
                file.write('''
from ...Generated.{0}Entity import {0}Entity as Entity
class {0}Entity(Entity):
    def __init__(self):
        super().__init__()'''.format(context.name))
                file.close()

    def generate_component(self):
        for key, context in self.contexts.items():
            self.render_mako("Components" , 'ecs_make_component.mako', context)

    def generate_matcher(self):
        for key, context in self.contexts.items():
            self.render_mako("Matchers" , 'ecs_matcher.mako', context)

    def generate_autoinc(self):
        template = Template(filename=str(self.mako_path / "ecs_autoinc.mako"),
                            module_directory=os.path.join(self.script_path, 'makoCache'))
        file_name = os.path.join(self.out_path, "Contexts.py" )
        file = utils.open_file(file_name, 'w')
        content = template.render(
            contexts=self.contexts,
            source_path=self.source
        )
        content = content.replace('\n', '')
        content = content.replace('\r\n', '')
        file.write(content)
        file.close()

    def copy_python_entitas_source(self):
        utils.copytree(self.python_entitas_soure_path, self.out_path / 'Source')
        return

    def generate_all_init(self):
        file_name = os.path.join(self.out_path / "../__init__.py")
        file = utils.open_file(file_name, 'w')
        file.write('from .Generated import *')
        entity_inc_list = []
        for key, context in self.contexts.items():
            entity_inc_list.append(context.name + 'Entity')
        file.write('''
from .Extension.Entity import {0}'''.format(','.join(entity_inc_list)))
        entity_inc_list = []
        for key, context in self.contexts.items():
            entity_inc_list.append(context.name + 'Context')
        file.write('''
from .Extension.Context import {0}'''.format(','.join(entity_inc_list)))
    def generate(self):
        self.generate_context()
        self.generate_component()
        self.generate_entity()
        self.generate_autoinc()
        self.copy_python_entitas_source()
        self.generate_all_init()




"""
entitas.entity
~~~~~~~~~~~~~~
An entity is a container holding data to represent certain
objects in your application. You can add, replace or remove data
from entities.

Those containers are called 'components'. They are represented by
namedtuples for readability.
"""

from .utils import Event
from .exceptions import (
    EntityNotEnabled, AlreadyAddedComponent, MissingComponent)


class Entity(object):
    """Use context.create_entity() to create a new entity and
    context.destroy_entity() to destroy it.
    You can add, replace and remove components to an entity.
    """

    def __init__(self):

        #: Occurs when a component gets added.
        self.on_component_added = Event()

        #: Occurs when a component gets removed.
        self.on_component_removed = Event()

        #: Occurs when a component gets replaced.
        self.on_component_replaced = Event()

        #: Dictionary mapping component type and component instance.
        self._components = {}

        #: Each entity has its own unique creationIndex which will be
        #: set by the context when you create the entity.
        self._creation_index = 0

        #: The context manages the state of an entity.
        #: Active entities are enabled, destroyed entities are not.
        self._is_enabled = False

    def activate(self, creation_index):
        self._creation_index = creation_index
        self._is_enabled = True

    def add(self, comp_type, *args):
        """Adds a component.
        :param comp_type: namedtuple type
        :param *args: (optional) data values
        """
        if not self._is_enabled:
            raise EntityNotEnabled(
                'Cannot add component {!r}: {} is not enabled.'
                .format(comp_type.__name__, self))

        if self.has(comp_type):
            raise AlreadyAddedComponent(
                'Cannot add another component {!r} to {}.'
                .format(comp_type.__name__, self))

        new_comp = comp_type._make(args)
        self._components[comp_type] = new_comp
        self.on_component_added(self, new_comp)

    def remove(self, comp_type):
        """Removes a component.
        :param comp_type: namedtuple type
        """
        if not self._is_enabled:
            raise EntityNotEnabled(
                'Cannot remove component {!r}: {} is not enabled.'
                .format(comp_type.__name__, self))

        if not self.has(comp_type):
            raise MissingComponent(
                'Cannot remove unexisting component {!r} from {}.'
                .format(comp_type.__name__, self))

        self._replace(comp_type, None)

    def replace(self, comp_type, *args):
        """Replaces an existing component or adds it if it doesn't exist
        yet.
        :param comp_type: namedtuple type
        :param *args: (optional) data values
        """
        if not self._is_enabled:
            raise EntityNotEnabled(
                'Cannot replace component {!r}: {} is not enabled.'
                .format(comp_type.__name__, self))

        if self.has(comp_type):
            self._replace(comp_type, args)
        else:
            self.add(comp_type, *args)

    def _replace(self, comp_type, args):
        previous_comp = self._components[comp_type]
        if args is None:
            del self._components[comp_type]
            self.on_component_removed(self, previous_comp)
        else:
            new_comp = comp_type._make(args)
            self._components[comp_type] = new_comp
            self.on_component_replaced(self, previous_comp, new_comp)

    def get(self, comp_type):
        """Retrieves a component by its type.
        :param comp_type: namedtuple type
        :rtype: namedtuple
        """
        if not self.has(comp_type):
            raise MissingComponent(
                'Cannot get unexisting component {!r} from {}.'
                .format(comp_type.__name__, self))

        return self._components[comp_type]

    def has(self, *args):
        """Checks if the entity has all components of the given type(s).
        :param args: namedtuple types
        :rtype: bool
        """
        if len(args) == 1:
            return args[0] in self._components

        return all([comp_type in self._components for comp_type in args])

    def has_any(self, *args):
        """Checks if the entity has any component of the given type(s).
        :param args: namedtuple types
        :rtype: bool
        """
        return any([comp_type in self._components for comp_type in args])

    def remove_all(self):
        """Removes all components."""
        for comp_type in list(self._components):
            self._replace(comp_type, None)

    def destroy(self):
        """This method is used internally. Don't call it yourself.
        Use context.destroy_entity(entity).
        """
        self._is_enabled = False
        self.remove_all()

    def __repr__(self):
        """ <Entity_0 [Position(x=1, y=2, z=3)]> """
        return '<Entity_{} [{}]>'.format(
            self._creation_index,
            ', '.join([str(self._components[x]) for x in self._components]))



from abc import ABCMeta, abstractmethod

from .collector import Collector


class InitializeProcessor(metaclass=ABCMeta):
    @abstractmethod
    def initialize(self):
        pass


class ExecuteProcessor(metaclass=ABCMeta):
    @abstractmethod
    def execute(self):
        pass


class CleanupProcessor(metaclass=ABCMeta):
    @abstractmethod
    def cleanup(self):
        pass


class TearDownProcessor(metaclass=ABCMeta):
    @abstractmethod
    def tear_down(self):
        pass


class ReactiveProcessor(ExecuteProcessor):

    def __init__(self, context):
        self._collector = self._get_collector(context)
        self._buffer = []

    @abstractmethod
    def get_trigger(self):
        pass

    @abstractmethod
    def filter(self, entity):
        pass

    @abstractmethod
    def react(self, entities):
        pass

    def activate(self):
        self._collector.activate()

    def deactivate(self):
        self._collector.deactivate()

    def clear(self):
        self._collector.clear_collected_entities()

    def execute(self):
        if self._collector.collected_entities:
            for entity in self._collector.collected_entities:
                if self.filter(entity):
                    self._buffer.append(entity)

            self._collector.clear_collected_entities()

            if self._buffer:
                self.react(self._buffer)
                self._buffer.clear()

    def _get_collector(self, context):
        trigger = self.get_trigger()
        collector = Collector()

        for matcher in trigger:
            group_event = trigger[matcher]
            group = context.get_group(matcher)
            collector.add(group, group_event)

        return collector


class Processors(InitializeProcessor, ExecuteProcessor,
                 CleanupProcessor, TearDownProcessor):

    def __init__(self):
        self._initialize_processors = []
        self._execute_processors = []
        self._cleanup_processors = []
        self._tear_down_processors = []

    def add(self, processor):
        if isinstance(processor, InitializeProcessor):
            self._initialize_processors.append(processor)

        if isinstance(processor, ExecuteProcessor):
            self._execute_processors.append(processor)

        if isinstance(processor, CleanupProcessor):
            self._cleanup_processors.append(processor)

        if isinstance(processor, TearDownProcessor):
            self._tear_down_processors.append(processor)

    def initialize(self):
        for processor in self._initialize_processors:
            processor.initialize()

    def execute(self):
        for processor in self._execute_processors:
            processor.execute()

    def cleanup(self):
        for processor in self._cleanup_processors:
            processor.cleanup()

    def tear_down(self):
        for processor in self._tear_down_processors:
            processor.tear_down()

    def activate_reactive_processors(self):
        for processor in self._execute_processors:
            if isinstance(processor, ReactiveProcessor):
                processor.activate()

            if isinstance(processor, Processors):
                processor.activate_reactive_processors()

    def deactivate_reactive_processors(self):
        for processor in self._execute_processors:
            if issubclass(processor, ReactiveProcessor):
                processor.deactivate()

            if isinstance(processor, Processors):
                processor.deactivate_reactive_processors()

    def clear_reactive_processors(self):
        for processor in self._execute_processors:
            if issubclass(processor, ReactiveProcessor):
                processor.clear()

            if isinstance(processor, Processors):
                processor.clear_reactive_processors()



def get_expr_repr(expr):
    return '' if expr is None else ','.join([x.__name__ for x in expr])


class Matcher(object):

    def __init__(self, *args, **kwargs):
        self._all = args if args else kwargs.get('all_of', None)
        self._any = kwargs.get('any_of', None)
        self._none = kwargs.get('none_of', None)

    def matches(self, entity):
        all_cond = self._all is None or entity.has(*self._all)
        any_cond = self._any is None or entity.has_any(*self._any)
        none_cond = self._none is None or not entity.has_any(*self._none)

        return all_cond and any_cond and none_cond

    def __repr__(self):
        return '<Matcher [all=({}) any=({}) none=({})]>'.format(
            get_expr_repr(self._all),
            get_expr_repr(self._any),
            get_expr_repr(self._none))



from collections import deque

from .entity import Entity
from .matcher import Matcher
from .group import Group
from .exceptions import MissingEntity


class Context(object):
    """A context is a data structure managing entities."""

    def __init__(self):

        #: Entities retained by this context.
        self._entities = set()

        #: An object pool to recycle entities.
        self._reusable_entities = deque()

        #: Entities counter.
        self._entity_index = 0

        #: Dictionary of matchers mapping groups.
        self._groups = {}

        self._entity_indices = {}

    @property
    def entities(self):
        return self._entities

    def has_entity(self, entity):
        """Checks if the context contains this entity.
        :param entity: Entity
        :rtype: bool
        """
        return entity in self._entities

    def _create_entity(self):
        assert False

    def create_entity(self):
        """Creates an entity. Pop one entity from the pool if it is not
        empty, otherwise creates a new one. Increments the entity index.
        Then adds the entity to the list.
        :rtype: Entity
        """
        entity = (self._reusable_entities.pop() if self._reusable_entities
                  else self._create_entity())

        entity.activate(self._entity_index)
        self._entity_index += 1

        self._entities.add(entity)

        entity.on_component_added += self._comp_added_or_removed
        entity.on_component_removed += self._comp_added_or_removed
        entity.on_component_replaced += self._comp_replaced

        return entity

    def destroy_entity(self, entity):
        """Removes an entity from the list and add it to the pool. If
        the context does not contain this entity, a
        :class:`MissingEntity` exception is raised.
        :param entity: Entity
        """
        if not self.has_entity(entity):
            raise MissingEntity()

        entity.destroy()

        self._entities.remove(entity)
        self._reusable_entities.append(entity)

    def get_group(self, matcher):
        """User can ask for a group of entities from the context. The
        group is identified through a :class:`Matcher`.
        :param entity: Matcher
        """
        if matcher in self._groups:
            return self._groups[matcher]

        group = Group(matcher)

        for entity in self._entities:
            group.handle_entity_silently(entity)

        self._groups[matcher] = group

        return group

    def set_unique_component(self, comp_type, *args):
        self.create_entity().add(comp_type, *args)

    def get_unique_component(self, comp_type):
        group = self.get_group(Matcher(comp_type))
        return group.single_entity.get(comp_type)

    def add_entity_index(self, entity_index):
        self._entity_indices[entity_index.type] = entity_index

    def get_entity_index(self, comp_type):
        return self._entity_indices[comp_type]

    def _comp_added_or_removed(self, entity, comp):
        for matcher in self._groups:
            self._groups[matcher].handle_entity(entity, comp)

    def _comp_replaced(self, entity, previous_comp, new_comp):
        for matcher in self._groups:
            group = self._groups[matcher]
            group.update_entity(entity, previous_comp, new_comp)

    def __repr__(self):
        return '<Context ({}/{})>'.format(
            len(self._entities), len(self._reusable_entities))



from abc import ABCMeta, abstractmethod

from .exceptions import EntitasException


class AbstractEntityIndex(metaclass=ABCMeta):

    def __init__(self, comp_type, group, *fields):
        self.type = comp_type
        self._group = group
        self._fields = fields
        self._index = {}
        self._activate()

    def __del__(self):
        self._deactivate()

    def _activate(self):
        self._group.on_entity_added += self._on_entity_added
        self._group.on_entity_removed += self._on_entity_removed
        self._index_entities()
        return self

    def _deactivate(self):
        self._group.on_entity_added -= self._on_entity_added
        self._group.on_entity_removed -= self._on_entity_removed
        self._index.clear()

    def _index_entities(self):
        for entity in self._group.entities:
            for field in self._fields:
                self._add_entity(getattr(entity.get(self.type), field), entity)

    def _on_entity_added(self, entity, component):
        for field in self._fields:
            self._add_entity(getattr(component, field), entity)

    def _on_entity_removed(self, entity, component):
        for field in self._fields:
            self._remove_entity(getattr(component, field), entity)

    @abstractmethod
    def _add_entity(self, key, entity):
        pass

    @abstractmethod
    def _remove_entity(self, key, entity):
        pass


class EntityIndex(AbstractEntityIndex):

    def get_entities(self, key):
        if key not in self._index:
            self._index[key] = set()
        return self._index[key]

    def _add_entity(self, key, entity):
        self.get_entities(key).add(entity)

    def _remove_entity(self, key, entity):
        self.get_entities(key).remove(entity)


class PrimaryEntityIndex(AbstractEntityIndex):

    def get_entity(self, key):
        return self._index[key]

    def _add_entity(self, key, entity):
        if key in self._index:
            raise EntitasException(
                "Entity for key '{key}' already exists!".format(key=key),
                "Only one entity for a primary key is allowed.")

        self._index[key] = entity

    def _remove_entity(self, key, entity):
        del self._index[key]



class Event(object):
    """C# events in Python."""

    def __init__(self):
        self._listeners = []

    def __call__(self, *args, **kwargs):
        for listener in self._listeners:
            listener(*args, **kwargs)

    def __add__(self, listener):
        if listener not in self._listeners:
            self._listeners.append(listener)
        return self

    def __sub__(self, listener):
        if listener in self._listeners:
            self._listeners.remove(listener)
        return self



from enum import Enum

from .utils import Event
from .exceptions import GroupSingleEntity


class GroupEvent(Enum):
    ADDED = 1
    REMOVED = 2
    ADDED_OR_REMOVED = 3


class Group(object):
    """Use context.get_group(matcher) to get a group of entities which
    match the specified matcher. Calling context.get_group(matcher) with
    the same matcher will always return the same instance of the group.

    The created group is managed by the context and will always be up to
    date. It will automatically add entities that match the matcher or
    remove entities as soon as they don't match the matcher anymore.
    """

    def __init__(self, matcher):

        #: Occurs when an entity gets added.
        self.on_entity_added = Event()

        #: Occurs when an entity gets removed.
        self.on_entity_removed = Event()

        #: Occurs when a component of an entity in the group gets
        # replaced.
        self.on_entity_updated = Event()

        self._matcher = matcher
        self._entities = set()

    @property
    def entities(self):
        return self._entities

    @property
    def single_entity(self):
        """Returns the only entity in this group.
        It will return None if the group is empty.
        It will throw a :class:`MissingComponent` if the group has more
        than one entity.
        """
        count = len(self._entities)

        if count == 1:
            return min(self._entities)
        if count == 0:
            return None

        raise GroupSingleEntity(
            'Cannot get a single entity from a group containing {} entities.',
            len(self._entities))

    def handle_entity_silently(self, entity):
        """This is used by the context to manage the group.
        :param matcher: Entity
        """
        if self._matcher.matches(entity):
            self._add_entity_silently(entity)
        else:
            self._remove_entity_silently(entity)

    def handle_entity(self, entity, component):
        """This is used by the context to manage the group.
        :param matcher: Entity
        """
        if self._matcher.matches(entity):
            self._add_entity(entity, component)
        else:
            self._remove_entity(entity, component)

    def update_entity(self, entity, previous_comp, new_comp):
        """This is used by the context to manage the group.
        :param matcher: Entity
        """
        if entity in self._entities:
            self.on_entity_removed(entity, previous_comp)
            self.on_entity_added(entity, new_comp)
            self.on_entity_updated(entity, previous_comp, new_comp)

    def _add_entity_silently(self, entity):
        if entity not in self._entities:
            self._entities.add(entity)
            return True
        return False

    def _add_entity(self, entity, component):
        entity_added = self._add_entity_silently(entity)
        if entity_added:
            self.on_entity_added(entity, component)

    def _remove_entity_silently(self, entity):
        if entity in self._entities:
            self._entities.remove(entity)
            return True
        return False

    def _remove_entity(self, entity, component):
        entity_removed = self._remove_entity_silently(entity)
        if entity_removed:
            self.on_entity_removed(entity, component)

    def __repr__(self):
        return '<Group [{}]>'.format(self._matcher)



import codecs
import os.path
import stat
import os
import os.path
import shutil
from pathlib import Path


def open_file(fullname, mode, encoding=None):
    if not os.path.exists(os.path.dirname(fullname)):
        os.makedirs(os.path.dirname(fullname), 0o777, True)

    if os.path.exists(fullname):
        os.chmod(fullname, stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
    return codecs.open(fullname, mode, encoding)


def get_file_list_with_suffix(list_dir, suffix, bRelationPath = False,recursion = True, appendPath = ""):
    ret_list = []
    list_dir = Path(list_dir)
    # list_dir = os.path.normpath(list_dir).replace("\\", "/")
    file_list = os.listdir(list_dir)
    for i in range(len(file_list)):
        full_path = list_dir / file_list[i]
        if recursion and os.path.isdir(full_path):
            ret_list.extend(get_file_list_with_suffix(full_path, suffix, bRelationPath, recursion, full_path.relative_to(list_dir)))
        if file_list[i].endswith(suffix):
            if bRelationPath:
                ret_list.append(str((Path(appendPath) / file_list[i])))
            else:
                ret_list.append(str(full_path))
    return ret_list


def get_map_with_list(list):
    out_map = {}
    for i in range(len(list)):
        if list[i] not in out_map:
            out_map[list[i]] = list[i]
    return out_map


def copytree(src, dst, symlinks=False):
    names = os.listdir(src)
    if not os.path.isdir(dst):
        os.makedirs(dst)

    errors = []
    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        try:
            if symlinks and os.path.islink(srcname):
                linkto = os.readlink(srcname)
                os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
                copytree(srcname, dstname, symlinks)
            else:
                if os.path.isdir(dstname):
                    os.rmdir(dstname)
                elif os.path.isfile(dstname):
                    os.remove(dstname)
                shutil.copy2(srcname, dstname)
            # XXX What about devices, sockets etc.?
        except (IOError, os.error) as why:
            errors.append((srcname, dstname, str(why)))
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except OSError as err:
            errors.extend(err.args[0])
    try:
        shutil.copystat(src, dst)
    except WindowsError:
        # can't copy file access times on Windows
        pass
    except OSError as why:
        errors.extend((src, dst, str(why)))
    if errors:
        print(errors)


def get_python_fiel_Path(__file__):
    return os.path.normpath(os.path.split(os.path.realpath(__file__))[0])


def get_file_name_without_ext(full_path):
    return os.path.split(full_path)[1].split('.')[0]


def print_error(msg):
    print(msg)



class EntityNotEnabled(Exception):
    """The entity is not enabled."""


class AlreadyAddedComponent(Exception):
    """The entity already contains this type of component."""


class MissingComponent(Exception):
    """The entity does not contain this type of component."""


class MissingEntity(Exception):
    """The context does not contain this entity."""


class GroupSingleEntity(Exception):
    """The group contains more than one entity."""


class EntitasException(Exception):
    def __init__(self, message, hint):
        super().__init__(message + '\n' + hint if hint else message)



# -*- coding: utf-8 -*-
from .BaseParser import BaseParser, Component
from . import utils


class PythonParser(BaseParser):
    def load_context_config(self):
        json_files = utils.get_file_list_with_suffix(self.component_path, '.lua')
        for file_path in json_files:
            if not file_path.endswith('entitas.lua'):
                with open(file_path, 'r', encoding='utf8') as load_f:
                    text = load_f.read()
                    ttt = self.lua.compile(text)()
                    components = self.get_components(ttt)
                    for comp in components:
                        for tag in comp.tag:
                            self.contexts[tag].addComponent(comp)

    def get_components(self, table):
        ret = []
        for key in table:
            ret.append(self.get_component(key, table[key]))
        return ret

    def get_component(self, key, table):
        comp = Component(key)
        if table.data:
            comp.set_data(list(table.data.values()))
        if table.single is not None:
            comp.set_single(True)
        comp.set_tag(list(table.tag.values()))
        if table.event is not None:
            comp.set_event(table.event)

        # if table.attr is not None:
        #     for at in table.attr:
        #         comp.add_attr(at, at)
        return comp

    def parse_context(self, key, table):
        context_files = utils.get_file_list_with_suffix(self.config_path, '.' + self.parser_tag)
        comp = Component(key)
        if table.data:
            comp.set_data(list(table.data.values()))
        if table.single is not None:
            comp.set_single(True)

        comp.set_tag(list(table.tag.values()))
        if table.event is not None:
            comp.set_event(table.event)

        # if table.attr is not None:
        #     for at in table.attr:
        #         comp.add_attr(at, at)
        return comp


from setuptools import setup, find_packages

if __name__ == "__main__":
    with open("README.md", "r") as fh:
        long_description = fh.read()
    setup(
        name="python_entitas",
        version="0.0.2",
        author="hangangliu",
        keywords=("entitas"),
        license="MIT Licence",
        author_email="hangangliurd@gmail.com",
        description="A small example package",
        long_description=long_description,
        url="https://github.com/UpUpLiu/python_entitas",
        packages=find_packages(),
        package_data={'': ['LICENSE']},
        include_package_data=True,
        platforms="any",
        classifiers=[
            "Programming Language :: Python :: 3",
        ],
        install_requires=[
            'Mako >= 1.0.7',
            'lupa >= 1.8'
        ]
    )





from .entity import Entity
from .entity_index import PrimaryEntityIndex, EntityIndex
from .context import Context
from .matcher import Matcher
from .group import Group, GroupEvent
from .collector import Collector
from .processors import (
    Processors, InitializeProcessor, ExecuteProcessor, CleanupProcessor,
    TearDownProcessor, ReactiveProcessor
)
from .utils import Event
from .exceptions import (
    AlreadyAddedComponent, MissingComponent, MissingEntity, GroupSingleEntity,
    EntitasException
)



from .group import GroupEvent


class Collector(object):

    def __init__(self):
        self._collected_entities = set()
        self._groups = {}

    @property
    def collected_entities(self):
        return self._collected_entities

    def add(self, group, group_event):
        self._groups[group] = group_event

    def activate(self):
        for group in self._groups:
            group_event = self._groups[group]

            added_event = group_event == GroupEvent.added
            removed_event = group_event == GroupEvent.removed
            added_or_removed_event = group_event == GroupEvent.added_or_removed

            if added_event or added_or_removed_event:
                group.on_entity_added -= self._add_entity
                group.on_entity_added += self._add_entity

            if removed_event or added_or_removed_event:
                group.on_entity_removed -= self._add_entity
                group.on_entity_removed += self._add_entity

    def deactivate(self):
        for group in self._groups:
            group.on_entity_added -= self._add_entity
            group.on_entity_removed -= self._add_entity

        self.clear_collected_entities()

    def clear_collected_entities(self):
        self._collected_entities.clear()

    def _add_entity(self, entity):  # , component
        self._collected_entities.add(entity)

    def __repr__(self):
        return '<Collector [{}]'.format(', '.join(map(str, self._groups)))
